
package Bundle::DBD::KB;

use strict;
use warnings;
use 5.008001;

our $VERSION = '3.18.4';

1;

__END__

=head1 NAME

Bundle::DBD::KB - A bundle to install all DBD::KB related modules

=head1 SYNOPSIS

C<perl -MCPAN -e 'install Bundle::DBD::KB'>

=head1 CONTENTS

DBI

DBD::KB

=head1 DESCRIPTION

This bundle includes all the modules needed for DBD::KB (the Perl 
interface to the Postgres database system). Please feel free to 
ask for help or report any problems to dbd-kb@perl.org.

=cut

=head1 AUTHOR

Greg Sabino Mullane E<lt>F<greg@turnstep.com>E<gt>

=cut

